/*
 *  obexftp/client.c: ObexFTP client library
 *
 *  Copyright (c) 2002 Christian W. Zuckschwerdt <zany@triq.net>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *     
 */

/* Funciones de apoyo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <gtk/gtk.h>

#include "global.h"

static const char* SDP_SERVICE_NAME = "OBEX Object Push";
static const char* SDP_SERVICE_PROVIDER = "obexpushd";
static const char* SDP_SERVICE_DESCR = "dummy description";

volatile int finished;	/* Para saber si ha terminado de hacer eventos || to know if it has finished making events */
volatile int last_rsp = OBEX_RSP_BAD_REQUEST;

int enviar(char MAC[20], char * fichero)
{
	int channel, ret, exitval = 1;
	obex_t *handle = NULL;
	obex_object_t *object;
	bdaddr_t bda;
	
	handle = inicia_obex();
	if (handle != NULL)
	{
		channel = obexftp_browse_bt(MAC, OBEX_OBJPUSH_SVCLASS_ID);
		
		if (channel == -1)
			return -1;
		
		g_print("Canal: %d\n",channel);
		str2ba(MAC, &bda);
		ret = BtOBEX_TransportConnect(handle, BDADDR_ANY, &bda, channel);
		if(ret<0)
			return -1;

		object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
		ret = do_sync_request(handle, object, 0);
	
		if ((last_rsp != OBEX_RSP_SUCCESS) || (ret < 0))
			return -2;
	
		if ((object = build_object_from_file(handle, fichero)))
		{
			ret = do_sync_request(handle, object, 0);
			if ((last_rsp != OBEX_RSP_SUCCESS) || (ret < 0))
				exitval = -3;
		}
		else
			exitval = -3;
	
		object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
		ret = do_sync_request(handle, object, 0);
		if ((last_rsp != OBEX_RSP_SUCCESS) || (ret < 0)) 
			exitval = -3;
		
		OBEX_Cleanup (handle);	//Cierro las conexiones || close connections
	}
	return exitval;
}

obex_t * inicia_obex()
{
	GtkWidget *dlgError;
	obex_t * handle=NULL;
	
	handle = OBEX_Init (OBEX_TRANS_BLUETOOTH, obex_event, 0);	// inicializa obex
	
	if (handle==NULL)	/* No encuentra dispositivo para enviar || there is not dispositive to send or recive*/
	{
		dlgError = gtk_message_dialog_new (NULL,
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_MESSAGE_ERROR,
                                  GTK_BUTTONS_CLOSE,
                                  "Error: Hubo un error al iniciar el dispositivo.");
 		gtk_dialog_run (GTK_DIALOG (dlgError));
 		gtk_widget_destroy (dlgError);
		g_print("[!] Error. Hubo un error al iniciar el dispositivo.\n An Error ocurred on init device");
		return NULL;
	}
	return handle;
}

void espera_conexiones()
{
	obex_t * handle=NULL;
	sdp_session_t* server_session;
	GtkWidget *dlgError;

	
	finished=0;
	
	if (busca_disp() < 0)	//Error
	{
		dlgError = gtk_message_dialog_new (NULL,
					  GTK_DIALOG_DESTROY_WITH_PARENT,
					  GTK_MESSAGE_ERROR,
					  GTK_BUTTONS_CLOSE,
					  "Error: El dispositivo local no esta disponible.");
		gtk_dialog_run (GTK_DIALOG (dlgError));
		gtk_widget_destroy (dlgError);
		g_print("[!] Error. Dispositivo Bluetooth local no disponible.\n The local device is not available.\n");
	}
	else
	{
		handle = inicia_obex();
		if (handle != NULL)
		{		
			server_session = bt_sdp_session_open (10);		/* Registro el canal 10 para escuchar || register channel 10 to listen */ 
			BtOBEX_ServerRegister(handle, NULL, 10);		// comienza a escuchar || begin to listen 
			
			while (!finished && !cancel) 	/* Mientras siga habiendo eventos a los que responder || while exist events*/
			{
				OBEX_HandleInput(handle, 1);
				
				/* Le dejo hacer los eventos || Let make gtk events */
				while (gtk_events_pending ())
					gtk_main_iteration ();
			}
		
			bt_sdp_session_close (server_session);	/* Cierro el canal 10 || close channel 10*/
			OBEX_Cleanup (handle);		//Cierro las conexiones || close connections
		}
	}
}

/*
 * Función safe_save_file ()
 *
 *   Guardamos el archivo en la carpeta personal
 *	  Save the file in personal folder
 */
int safe_save_file(char *name, const uint8_t *buf, int len)
{
	int fd;
	int actual;
	char *ruta=NULL;
	char aux[200];
	
	ruta=getenv("HOME"); //Obtengo la ruta de la carpeta personal || get the route to /home
	strcpy(aux,ruta);
    strcat(aux,"/");	//Le añadimos una / || add "/"
	strcat(aux,name);	//Y le concatenamos el nombre del archivo || And concatenated the name of file
	
	fd = open(aux, O_RDWR | O_CREAT, DEFFILEMODE);

	if ( fd < 0) {
		perror( aux);
		return -1;
	}
	
	actual = write(fd, buf, len);
	close(fd);

	g_print("Guardado %s (%d bytes)\n", aux, actual);

	return actual;
}

/* When we got a PUT */

void put_done(obex_object_t *object, obex_t *handle)
{
	obex_headerdata_t hv;
	uint8_t hi;
	int hlen;
	const uint8_t *body = NULL;
	int body_len = 0;
	char *name = NULL;
	char *namebuf = NULL;

	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen))
	{
		switch(hi)
		{
			case OBEX_HDR_BODY:
				body = hv.bs;
				body_len = hlen;
				break;

			case OBEX_HDR_NAME:
				if( (namebuf = malloc(hlen / 2)))
				{
					OBEX_UnicodeToChar(namebuf, hv.bs, hlen);
					name = namebuf;
				}
				break;

			case OBEX_HDR_LENGTH:
				g_print("\nsize = %d bytes\n", hv.bq4);
				break;

			case OBEX_HDR_TYPE:
				g_print("type = %s\n", hv.bs);
				break;

			case OBEX_HDR_TIME:
				g_print("date = %s\n", hv.bs);
				break;

			case HEADER_CREATOR_ID:
				g_print("creator ID = 0x%x\n", hv.bq4);
				break;

			default:
				g_print("%s(): skipped header 0x%02x\n", __FUNCTION__, hi);
		}
	}

	if(!body)  
	{
		g_print("%s(): got a PUT without a body\n", __FUNCTION__);
		return;
	}

	if(!name)  
	{
		name = "unknown_object.bin";
		g_print("-- warning: no filename received; changing to %s\n", name);
	}

	safe_save_file(name, body, body_len);
	free(namebuf);
}



/*
 * Se le llama cuando una peticion va a llegar o ya ha llegado
 * Se le llama en OBEX_EV_REQ
 *
 * When  a request is comming or it has arrived
 */
void server_request(obex_object_t *object, obex_t *handle, int event, int cmd)
{	
	switch(cmd)  
	{
		case OBEX_CMD_SETPATH:
			g_print("Received SETPATH command\n");
			OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			break;

		case OBEX_CMD_PUT:
			OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			put_done(object, handle);
			break;

		case OBEX_CMD_CONNECT:
			OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
			break;

		case OBEX_CMD_DISCONNECT:
			OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
			break;

		default:
			g_print("%s() Denied %02x request\n", __FUNCTION__, cmd);
			OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
			break;
	}
	return;
}

/*
 * Funcion obex_event ()
 *
 *    Llamadas cuando ocurre un evento || when ocurred an event
 *
 */
void obex_event(obex_t *handle, obex_object_t *object, int mode, int event, int obex_cmd, int obex_rsp)
{
	switch(event)
	{
		case OBEX_EV_PROGRESS:	/* here we can increment the progress bar */
			g_print(".");
			fflush(stdout);
			break;

		case OBEX_EV_REQDONE:
			if(mode == OBEX_CLIENT)
			{
				last_rsp = obex_rsp;
				finished = 1;
			}
			else
				if(obex_cmd==OBEX_CMD_DISCONNECT) 
				{
					finished = 1;
				}
			break;

		case OBEX_EV_REQHINT:
			/* Comes BEFORE the lib parses anything. */
			switch(obex_cmd) 
			{
				case OBEX_CMD_PUT:
				case OBEX_CMD_CONNECT:
				case OBEX_CMD_DISCONNECT:
					OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
					break;
				default:
					OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED,
					OBEX_RSP_NOT_IMPLEMENTED);
					break;
			}
			break;

		case OBEX_EV_REQ:
			// Comes when a server-request has been received.
			server_request(object, handle, event, obex_cmd);
			break;

		case OBEX_EV_LINKERR:
			g_print("Link broken (this does not have to be an error)!\n");
			finished = 1;
			break;

		default:
			g_print("Unknown event 0x%x!\n", event);
			break;
	}
}

/* Execute an OBEX-request synchronously. */
int do_sync_request(obex_t *handle, obex_object_t *object, int async)
{
	int ret;

	OBEX_Request(handle, object);
	ret = wait_for_rsp(handle);
	finished = 0;

	return ret;
}

obex_object_t *build_object_from_file(obex_t *handle, const char *filename)
{
	obex_headerdata_t hdd;
	uint8_t unicode_buf[200];
	int namebuf_len;
	obex_object_t *object;
	int file_size;
	uint8_t *buf;

	buf = easy_readfile(filename, &file_size);
	if(buf == NULL) 
		return NULL;

	object = OBEX_ObjectNew(handle, OBEX_CMD_PUT);

	namebuf_len = OBEX_CharToUnicode(unicode_buf, filename, sizeof(unicode_buf));

	hdd.bs = unicode_buf;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME,
	    hdd, namebuf_len, 0);

	hdd.bq4 = file_size;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_LENGTH,
	    hdd, sizeof(uint32_t), 0);

	hdd.bs = buf;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
	    hdd, file_size, 0);

	free(buf);
	return object;
}

sdp_session_t* bt_sdp_session_open (uint8_t channel) 
{
	sdp_session_t* session;

  	uuid_t uuid_srv;
	uuid_t uuid_group;
	uuid_t uuid_l2cap;
	uuid_t uuid_rfcomm;
	uuid_t uuid_obex;

	sdp_profile_desc_t desc_srv;
	sdp_data_t* chan;

	sdp_list_t* list_class = NULL;
	sdp_list_t* list_srv = NULL;
	sdp_list_t* list_group = NULL;
	sdp_list_t* list_l2cap = NULL;
	sdp_list_t* list_rfcomm = NULL;
	sdp_list_t* list_obex = NULL;

	uint8_t formats[] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0xFF };
	uint8_t fdtd = SDP_UINT8;
	void* dtds[sizeof(formats)];
	void* values[sizeof(formats)];
	sdp_data_t* list_formats = NULL;
	size_t i;

	sdp_list_t* list_proto = NULL;
	sdp_list_t* list_access = NULL;
	sdp_record_t* rec = sdp_record_alloc();
	int status;

	/* Service: OBEX-Push */
	list_class = sdp_list_append(list_class,sdp_uuid16_create(&uuid_srv,OBEX_OBJPUSH_SVCLASS_ID));
	sdp_set_service_classes(rec,list_class);

	/* the service is publicly browsable */
	list_group = sdp_list_append(list_group,sdp_uuid16_create(&uuid_group,PUBLIC_BROWSE_GROUP));
	sdp_set_browse_groups(rec,list_group);

	/* protocol descriptor list */
	list_l2cap = sdp_list_append(list_l2cap,sdp_uuid16_create(&uuid_l2cap,L2CAP_UUID));
	list_rfcomm = sdp_list_append(list_rfcomm,sdp_uuid16_create(&uuid_rfcomm,RFCOMM_UUID));
	list_obex = sdp_list_append(list_obex,sdp_uuid16_create(&uuid_obex,OBEX_UUID));

	chan = sdp_data_alloc(SDP_UINT8,&channel);
	list_rfcomm = sdp_list_append(list_rfcomm,chan);

	list_proto = sdp_list_append(list_proto,list_l2cap);
	list_proto = sdp_list_append(list_proto,list_rfcomm);
	list_proto = sdp_list_append(list_proto,list_obex);

	list_access = sdp_list_append(list_access,list_proto);
	sdp_set_access_protos(rec,list_access);

	desc_srv.version = 0x0100;
	list_srv = sdp_list_append(list_srv,sdp_uuid16_create(&desc_srv.uuid, OBEX_OBJPUSH_PROFILE_ID));
	sdp_set_profile_descs(rec,list_srv);

	for (i = 0; i < sizeof(formats); ++i)
	{
    	dtds[i] = &fdtd;
    	values[i] = formats+i;
  	}
	list_formats = sdp_seq_alloc(dtds,values,sizeof(formats));
	sdp_attr_add(rec,SDP_ATTR_SUPPORTED_FORMATS_LIST,list_formats);

	sdp_set_info_attr(rec,SDP_SERVICE_NAME,SDP_SERVICE_PROVIDER,SDP_SERVICE_DESCR);

	session = sdp_connect(BDADDR_ANY,BDADDR_LOCAL,SDP_RETRY_IF_BUSY);
	status = sdp_record_register(session,rec,0);
	if (status < 0)
	{
		bt_sdp_session_close(session);
		session = NULL;
	}

	sdp_data_free(chan);
	sdp_list_free(list_class,0);
	sdp_list_free(list_srv,0);
	sdp_list_free(list_group,0);
	sdp_list_free(list_l2cap,0);
	sdp_list_free(list_rfcomm,0);
	sdp_list_free(list_obex,0);
	sdp_data_free(list_formats);
	sdp_list_free(list_proto,0);
	sdp_list_free(list_access,0);
	/*   sdp_record_free(rec); */

	return session;
}

/* Cierra la sesion sdp. Deja de estar activo el canal de escucha || Close session sdp and now the channel to listen is not available */
int bt_sdp_session_close (sdp_session_t* session)
{
	return sdp_close(session);
}

/* Read a file and alloc a buffer for it */

uint8_t* easy_readfile(const char *filename, int *file_size)
{
	int actual;
	int fd;
	uint8_t *buf;

	*file_size = get_filesize(filename);
	g_print("%s (%d bytes)\n", filename, *file_size);

	fd = open(filename, O_RDONLY, 0);
	if (fd == -1) 
		return NULL;

	if(! (buf = malloc(*file_size)) )
		return NULL;

	actual = read(fd, buf, *file_size);
	close(fd); 

	*file_size = actual;
	return buf;
}

int get_filesize(const char *filename)
{
	struct stat stats;

	stat(filename, &stats);
	return (int) stats.st_size;
}

// Wait for a request to finish! */

int wait_for_rsp(obex_t *handle)
{
	int ret;
	
	while(!finished && !cancel)
	{
		/* Le dejo hacer los eventos || Let make gtk events */
		while (gtk_events_pending ())
        	gtk_main_iteration ();

		ret = OBEX_HandleInput(handle, 1);
		
		if(ret<0) 
			return ret;
	}

	return last_rsp;
}

int busca_disp ()
{
	int id_dispo;
	
	id_dispo = hci_get_route(NULL);
	
	return id_dispo;
}
